@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2017 Christian Geier et al.
+# Copyright (c) 2013-2022 khal contributors
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -27,35 +27,40 @@ import logging
 import re
 from calendar import isleap
 from time import strptime
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union
 
 import pytz
-from khal.exceptions import FatalError, DateTimeParseError
+
+from khal.exceptions import DateTimeParseError, FatalError
+
+from .custom_types import LocaleConfiguration, RRuleMapType
 
 logger = logging.getLogger('khal')
 
 
-def timefstr(dtime_list, timeformat):
+def timefstr(dtime_list: List[str], timeformat: str) -> dt.datetime:
     """converts the first item of a list (a time as a string) to a datetimeobject
 
     where the date is today and the time is given by a string
-    removes "used" elements of list
-
-    :type dtime_list: list(str)
-    :type timeformat: str
-    :rtype: datetime.datetime
+    removes "used" elements of list.
     """
     if len(dtime_list) == 0:
         raise ValueError()
-    time_start = dt.datetime.strptime(dtime_list[0], timeformat)
-    time_start = dt.time(*time_start.timetuple()[3:5])
+    datetime_start = dt.datetime.strptime(dtime_list[0], timeformat)
+    time_start = dt.time(*datetime_start.timetuple()[3:5])
     day_start = dt.date.today()
     dtstart = dt.datetime.combine(day_start, time_start)
     dtime_list.pop(0)
     return dtstart
 
 
-def datetimefstr(dtime_list, dateformat, default_day=None, infer_year=True,
-                 in_future=True):
+def datetimefstr(
+    dtime_list: List[str],
+    dateformat: str,
+    default_day: Optional[dt.date]=None,
+    infer_year: bool=True,
+    in_future: bool=True,
+) -> dt.datetime:
     """converts a datetime (as one or several string elements of a list) to
     a datetimeobject, if infer_year is True, use the `default_day`'s year as
     the year of the return datetimeobject,
@@ -74,32 +79,33 @@ def datetimefstr(dtime_list, dateformat, default_day=None, infer_year=True,
     parts = dateformat.count(' ') + 1
     dtstring = ' '.join(dtime_list[0:parts])
     # only time.strptime can parse the 29th of Feb. if no year is given
-    dtstart = strptime(dtstring, dateformat)
-    if infer_year and dtstart.tm_mon == 2 and dtstart.tm_mday == 29 and \
+    dtstart_struct = strptime(dtstring, dateformat)
+    if infer_year and dtstart_struct.tm_mon == 2 and dtstart_struct.tm_mday == 29 and \
             not isleap(default_day.year):
         raise ValueError
 
     for _ in range(parts):
-        dtime_list.pop(0)
+        item = dtime_list.pop(0)
+        if ' ' in item:
+            logger.warn('detected a space in datetime specification, this can lead to errors.')
+            logger.warn('Make sure not to quote your datetime specification.')
 
     if infer_year:
-        dtstart = dt.datetime(*(default_day.timetuple()[:1] + dtstart[1:5]))
+        dtstart = dt.datetime(*(default_day.timetuple()[:1] + dtstart_struct[1:5]))
         if in_future and dtstart < now:
             dtstart = dtstart.replace(year=dtstart.year + 1)
         if dtstart.date() < default_day:
             dtstart = dtstart.replace(year=default_day.year + 1)
         return dtstart
     else:
-        return dt.datetime(*dtstart[:5])
+        return dt.datetime(*dtstart_struct[:5])
 
 
-def weekdaypstr(dayname):
+def weekdaypstr(dayname: str) -> int:
     """converts an (abbreviated) dayname to a number (mon=0, sun=6)
 
     :param dayname: name of abbreviation of the day
-    :type dayname: str
     :return: number of the day in a week
-    :rtype: int
     """
 
     if dayname in ['monday', 'mon']:
@@ -119,7 +125,7 @@ def weekdaypstr(dayname):
     raise ValueError('invalid weekday name `%s`' % dayname)
 
 
-def construct_daynames(date_):
+def construct_daynames(date_: dt.date) -> str:
     """converts datetime.date into a string description
 
     either `Today`, `Tomorrow` or name of weekday.
@@ -132,13 +138,11 @@ def construct_daynames(date_):
         return date_.strftime('%A')
 
 
-def calc_day(dayname):
+def calc_day(dayname: str) -> dt.datetime:
     """converts a relative date's description to a datetime object
 
     :param dayname: relative day name (like 'today' or 'monday')
-    :type dayname: str
     :returns: date
-    :rtype: datetime.datetime
     """
     today = dt.datetime.combine(dt.date.today(), dt.time.min)
     dayname = dayname.lower()
@@ -156,15 +160,14 @@ def calc_day(dayname):
     return day
 
 
-def datefstr_weekday(dtime_list, _, **kwargs):
+def datefstr_weekday(dtime_list: List[str], timeformat: str, infer_year: bool) -> dt.datetime:
     """interprets first element of a list as a relative date and removes that
     element
 
     :param dtime_list: event description in list form
-    :type dtime_list: list
-    :returns: date
-    :rtype: datetime.datetime
+    :param timeformat: only here for compat reasons (having the same function signature)
 
+    :returns: date
     """
     if len(dtime_list) == 0:
         raise ValueError()
@@ -173,7 +176,10 @@ def datefstr_weekday(dtime_list, _, **kwargs):
     return day
 
 
-def datetimefstr_weekday(dtime_list, timeformat, **kwargs):
+def datetimefstr_weekday(dtime_list: List[str], timeformat: str, infer_year: bool) -> dt.datetime:
+    """
+    :param infer_year: only here for compat reasons (having the same function signature)
+    """
     if len(dtime_list) == 0:
         raise ValueError()
     day = calc_day(dtime_list[0])
@@ -184,39 +190,42 @@ def datetimefstr_weekday(dtime_list, timeformat, **kwargs):
     return dtime
 
 
-def guessdatetimefstr(dtime_list, locale, default_day=None, in_future=True):
+def guessdatetimefstr(
+    dtime_list: List[str],
+    locale: LocaleConfiguration,
+    default_day: Optional[dt.date]=None,
+    in_future=True,
+) -> Tuple[dt.datetime, bool]:
     """
-    :type dtime_list: list
-    :type locale: dict
-    :type default_day: datetime.datetime
     :param in_future: if set, shortdate(time) events will be set in the future
-    :type in_future: bool
-    :rtype: datetime.datetime
     """
     # if now() is called as default param, mocking with freezegun won't work
-    if default_day is None:
-        default_day = dt.datetime.now().date()
+    day = default_day or dt.datetime.now().date()
     # TODO rename in guessdatetimefstrLIST or something saner altogether
 
-    def timefstr_day(dtime_list, timeformat, **kwargs):
+    def timefstr_day(dtime_list: List[str], timeformat: str, infer_year: bool) -> dt.datetime:
         if locale['timeformat'] == '%H:%M' and dtime_list[0] == '24:00':
-            a_date = dt.datetime.combine(default_day, dt.time(0))
+            a_date = dt.datetime.combine(day, dt.time(0))
             dtime_list.pop(0)
         else:
             a_date = timefstr(dtime_list, timeformat)
-            a_date = dt.datetime(*(default_day.timetuple()[:3] + a_date.timetuple()[3:5]))
+            a_date = dt.datetime(*(day.timetuple()[:3] + a_date.timetuple()[3:5]))
         return a_date
 
-    def datetimefwords(dtime_list, _, **kwargs):
+    def datetimefwords(dtime_list: List[str], _: str, infer_year: bool) -> dt.datetime:
         if len(dtime_list) > 0 and dtime_list[0].lower() == 'now':
             dtime_list.pop(0)
             return dt.datetime.now()
         raise ValueError
 
-    def datefstr_year(dtime_list, dtformat, infer_year):
-        return datetimefstr(dtime_list, dtformat, default_day, infer_year, in_future)
+    def datefstr_year(dtime_list: List[str], dtformat: str, infer_year: bool) -> dt.datetime:
+        return datetimefstr(dtime_list, dtformat, day, infer_year, in_future)
 
     dtstart = None
+    fun: Callable[[List[str], str, bool], dt.datetime]
+    dtformat: str
+    all_day: bool
+    infer_year: bool
     for fun, dtformat, all_day, infer_year in [
             (datefstr_year, locale['datetimeformat'], False, True),
             (datefstr_year, locale['longdatetimeformat'], False, False),
@@ -224,8 +233,8 @@ def guessdatetimefstr(dtime_list, locale, default_day=None, in_future=True):
             (datetimefstr_weekday, locale['timeformat'], False, False),
             (datefstr_year, locale['dateformat'], True, True),
             (datefstr_year, locale['longdateformat'], True, False),
-            (datefstr_weekday, None, True, False),
-            (datetimefwords, None, False, False),
+            (datefstr_weekday, '', True, False),
+            (datetimefwords, '', False, False),
     ]:
         # if a `short` format contains a year, treat it as a `long` format
         if infer_year and '97' in dt.datetime(1997, 10, 11).strftime(dtformat):
@@ -237,15 +246,14 @@ def guessdatetimefstr(dtime_list, locale, default_day=None, in_future=True):
         else:
             return dtstart, all_day
     raise DateTimeParseError(
-        "Could not parse \"{}\".\nPlease check your configuration or run "
-        "`khal printformats` to see if this does match your configured "
+        f"Could not parse \"{dtime_list}\".\nPlease check your configuration "
+        "or run `khal printformats` to see if this does match your configured "
         "[long](date|time|datetime)format.\nIf you suspect a bug, please "
         "file an issue at https://github.com/pimutils/khal/issues/ "
-        "".format(dtime_list)
     )
 
 
-def timedelta2str(delta):
+def timedelta2str(delta: dt.timedelta) -> str:
     # we deliberately ignore any subsecond deltas
     total_seconds = int(abs(delta).total_seconds())
 
@@ -275,7 +283,7 @@ def timedelta2str(delta):
     return ' '.join(s)
 
 
-def guesstimedeltafstr(delta_string):
+def guesstimedeltafstr(delta_string: str) -> dt.timedelta:
     """parses a timedelta from a string
 
     :param delta_string: string encoding time-delta, e.g. '1h 15m'
@@ -285,8 +293,7 @@ def guesstimedeltafstr(delta_string):
 
     tups = re.split(r'(-?\d+)', delta_string)
     if not re.match(r'^\s*$', tups[0]):
-        raise ValueError('Invalid beginning of timedelta string "%s": "%s"'
-                         % (delta_string, tups[0]))
+        raise ValueError(f'Invalid beginning of timedelta string "{delta_string}": "{tups[0]}"')
     tups = tups[1:]
     res = dt.timedelta()
 
@@ -295,7 +302,7 @@ def guesstimedeltafstr(delta_string):
             numint = int(num)
         except ValueError:
             raise DateTimeParseError(
-                'Invalid number in timedelta string "%s": "%s"' % (delta_string, num))
+                f'Invalid number in timedelta string "{delta_string}": "{num}"')
 
         ulower = unit.lower().strip()
         if ulower == 'd' or ulower == 'day' or ulower == 'days':
@@ -309,17 +316,17 @@ def guesstimedeltafstr(delta_string):
               ulower == 'sec'):
             res += dt.timedelta(seconds=numint)
         else:
-            raise ValueError('Invalid unit in timedelta string "%s": "%s"'
-                             % (delta_string, unit))
+            raise ValueError(f'Invalid unit in timedelta string "{delta_string}": "{unit}"')
 
     return res
 
 
-def guessrangefstr(daterange, locale,
-                   default_timedelta_date=dt.timedelta(days=1),
-                   default_timedelta_datetime=dt.timedelta(hours=1),
-                   adjust_reasonably=False,
-                   ):
+def guessrangefstr(daterange: Union[str, List[str]],
+                   locale: LocaleConfiguration,
+                   default_timedelta_date: dt.timedelta=dt.timedelta(days=1),
+                   default_timedelta_datetime: dt.timedelta=dt.timedelta(hours=1),
+                   adjust_reasonably: bool=False,
+                   ) -> Tuple[dt.datetime, dt.datetime, bool]:
     """parses a range string
 
     :param daterange: date1 [date2 | timedelta]
@@ -328,44 +335,43 @@ def guessrangefstr(daterange, locale,
     :returns: start and end of the date(time) range  and if
         this is an all-day time range or not,
         **NOTE**: the end is *exclusive* if this is an allday event
-    :rtype: (datetime, datetime, bool)
-
     """
     range_list = daterange
     if isinstance(daterange, str):
         range_list = daterange.split(' ')
+        assert isinstance(range_list, list)
 
     if range_list == ['week']:
         today_weekday = dt.datetime.today().weekday()
-        start = dt.datetime.today() - dt.timedelta(days=(today_weekday - locale['firstweekday']))
-        end = start + dt.timedelta(days=8)
-        return start, end, True
+        startdt = dt.datetime.today() - dt.timedelta(days=(today_weekday - locale['firstweekday']))
+        enddt = startdt + dt.timedelta(days=8)
+        return startdt, enddt, True
 
     for i in reversed(range(1, len(range_list) + 1)):
-        start = ' '.join(range_list[:i])
-        end = ' '.join(range_list[i:])
+        startstr = ' '.join(range_list[:i])
+        endstr = ' '.join(range_list[i:])
         allday = False
         try:
             # figuring out start
-            split = start.split(" ")
+            split = startstr.split(" ")
             start, allday = guessdatetimefstr(split, locale)
             if len(split) != 0:
                 continue
 
             # and end
-            if len(end) == 0:
+            if len(endstr) == 0:
                 if allday:
                     end = start + default_timedelta_date
                 else:
                     end = start + default_timedelta_datetime
-            elif end.lower() == 'eod':
+            elif endstr.lower() == 'eod':
                 end = dt.datetime.combine(start.date(), dt.time.max)
-            elif end.lower() == 'week':
+            elif endstr.lower() == 'week':
                 start -= dt.timedelta(days=(start.weekday() - locale['firstweekday']))
                 end = start + dt.timedelta(days=8)
             else:
                 try:
-                    delta = guesstimedeltafstr(end)
+                    delta = guesstimedeltafstr(endstr)
                     if allday and delta.total_seconds() % (3600 * 24):
                         # TODO better error class, no logging in here
                         logger.fatal(
@@ -380,7 +386,7 @@ def guessrangefstr(daterange, locale,
 
                     end = start + delta
                 except (ValueError, DateTimeParseError):
-                    split = end.split(" ")
+                    split = endstr.split(" ")
                     end, end_allday = guessdatetimefstr(
                         split, locale, default_day=start.date(), in_future=False)
                     if len(split) != 0:
@@ -407,20 +413,28 @@ def guessrangefstr(daterange, locale,
             pass
 
     raise DateTimeParseError(
-        "Could not parse \"{}\".\nPlease check your configuration or run "
-        "`khal printformats` to see if this does match your configured "
+        f"Could not parse \"{daterange}\".\nPlease check your configuration or "
+        "run `khal printformats` to see if this does match your configured "
         "[long](date|time|datetime)format.\nIf you suspect a bug, please "
         "file an issue at https://github.com/pimutils/khal/issues/ "
-        "".format(daterange)
     )
 
 
-def rrulefstr(repeat, until, locale):
+def rrulefstr(repeat: str,
+              until: str,
+              locale: LocaleConfiguration,
+              timezone: Optional[dt.tzinfo],
+              ) -> RRuleMapType:
     if repeat in ["daily", "weekly", "monthly", "yearly"]:
-        rrule_settings = {'freq': repeat}
+        rrule_settings: RRuleMapType = {'freq': repeat}
         if until:
-            until_dt, is_date = guessdatetimefstr(until.split(' '), locale)
-            rrule_settings['until'] = until_dt
+            until_dt, _ = guessdatetimefstr(until.split(' '), locale)
+            if timezone:
+                rrule_settings['until'] = until_dt.\
+                    replace(tzinfo=timezone).\
+                    astimezone(pytz.UTC)
+            else:
+                rrule_settings['until'] = until_dt
         return rrule_settings
     else:
         logger.fatal("Invalid value for the repeat option. \
@@ -428,20 +442,15 @@ def rrulefstr(repeat, until, locale):
         raise FatalError()
 
 
-def eventinfofstr(info_string, locale, default_event_duration, default_dayevent_duration,
-                  adjust_reasonably=False, localize=False):
+def eventinfofstr(info_string: str,
+                  locale: LocaleConfiguration,
+                  default_event_duration: dt.timedelta,
+                  default_dayevent_duration: dt.timedelta,
+                  adjust_reasonably: bool=False,
+                  ) -> Dict[str, Any]:
     """parses a string of the form START [END | DELTA] [TIMEZONE] [SUMMARY] [::
     DESCRIPTION] into a dictionary with keys: dtstart, dtend, timezone, allday,
     summary, description
-
-    :param info_string:
-    :type info_string: string fitting the form
-    :param locale:
-    :type locale: locale
-    :param adjust_reasonably:
-    :type adjust_reasonably: passed on to guessrangefstr
-    :rtype: dictionary
-
     """
     description = None
     if " :: " in info_string:
@@ -449,10 +458,10 @@ def eventinfofstr(info_string, locale, default_event_duration, default_dayevent_
 
     parts = info_string.split(' ')
     summary = None
-    start = None
-    end = None
-    tz = None
-    allday = False
+    start: Optional[Union[dt.datetime, dt.date]] = None
+    end: Optional[Union[dt.datetime, dt.date]] = None
+    tz: Optional[pytz.BaseTzInfo] = None
+    allday: bool = False
     for i in reversed(range(1, len(parts) + 1)):
         try:
             start, end, allday = guessrangefstr(
@@ -475,21 +484,23 @@ def eventinfofstr(info_string, locale, default_event_duration, default_dayevent_
 
     if start is None or end is None:
         raise DateTimeParseError(
-            "Could not parse \"{}\".\nPlease check your configuration or run "
-            "`khal printformats` to see if this does match your configured "
-            "[long](date|time|datetime)format.\nIf you suspect a bug, please "
-            "file an issue at https://github.com/pimutils/khal/issues/ "
-            "".format(info_string)
+            f"Could not parse \"{info_string}\".\nPlease check your "
+            "configuration or run `khal printformats` to see if this does "
+            "match your configured [long](date|time|datetime)format.\nIf you "
+            "suspect a bug, please file an issue at "
+            "https://github.com/pimutils/khal/issues/ "
         )
 
     if tz is None:
         tz = locale['default_timezone']
 
     if allday:
+        assert isinstance(start, dt.datetime)
+        assert isinstance(end, dt.datetime)
         start = start.date()
         end = end.date()
 
-    info = {}
+    info: Dict[str, Any] = {}
     info["dtstart"] = start
     info["dtend"] = end
     info["summary"] = summary if summary else None
